// code by jph

// from 'Numerical Recipes in C++' 3rd edition by Press, Teukolsky, Vetterling, Flannery:
//  SVD, EIG, NEWTON, ...

#define     TENSOR_EIG            1E-15
#define     TENSOR_LU             1E-10
#define     TENSOR_QR             1E-06
#define     TENSOR_SCHUR          1E-06
#define     TENSOR_SVD            30

#define     TENSOR_NEWTON_CHECK   0
#define     TENSOR_NEWTON_CONT    1


template<class type>
tensor<type> tensor<type>::solve(tensor<type> rhs) {
  assert(ndims==2);
  assert(size[0]==size[1]);
  assert(size[1]==rhs.size[0]);
  int c0,c1,c2,c1s=0,k,swap;
  type tfac,tmax,tabs;
  type* copy=new type[numel];
  for (c0=0;c0<numel;++c0)
    copy[c0]=data[c0];
  tensor<type> sol(2,rhs.size[0],(rhs.ndims<=1)?1:rhs.size[1]);
  tensor<int> ind=tensor<int>::index(1,size[0]);
  valid=1;
  for (c0=0;c0<size[0];++c0) {
    k=c0;
    tmax=0;
    for (c1=c0;c1<size[0];++c1)
      if (tmax<(tabs=abs(copy[ind.data[c1]+c1s]))) {
        tmax=tabs;
        k=c1;
      }
    swap=ind.data[k];
    ind.data[k]=ind.data[c0];
    ind.data[c0]=swap;
    for (c1=c0+1;c1<size[0];++c1) {
      valid&=TENSOR_NO_DIVISION<abs(copy[ind.data[c0]+c1s]);
      if (!valid)
        break;
      tabs=copy[ind.data[c1]+c1s]/copy[ind.data[c0]+c1s]; // division
      for (c2=(c0+1)*size[0];c2<numel;c2+=size[0])
        copy[ind.data[c1]+c2]-=copy[ind.data[c0]+c2]*tabs;
      for (c2=0;c2<rhs.numel;c2+=size[0])
        rhs.data[ind.data[c1]+c2]-=rhs.data[ind.data[c0]+c2]*tabs;
    }
    c1s+=size[0];
  }
  for (c0=size[0]-1;c0>=0;--c0) {
    assert(TENSOR_NO_DIVISION<abs(copy[ind.data[c0]+c0*size[0]]));
    if (abs(copy[ind.data[c0]+c0*size[0]])<TENSOR_NO_DIVISION) {
      valid=0;
      break;
    }
    tfac=1/copy[ind.data[c0]+c0*size[0]]; // division
    for (c1=0;c1<sol.size[1];++c1) {
      tabs=rhs.data[ind.data[c0]+c1*size[0]];
      for (c2=c0+1;c2<size[0];++c2)
        tabs-=copy[ind.data[c0]+c2*size[0]]*sol.data[c2+c1*size[0]];
      sol.data[c0+c1*size[0]]=tabs*tfac;
    }
  }
  sol.ndims=rhs.ndims;
  delete[] copy;
  return sol;
}

template<class type>
type tensor<type>::det(void) { // TODO det() crashes for |this|==1
  assert(size[0]==size[1]);
  tensor<int> ind;
  type sgn;
  tensor<type> tmp=lu(ind,sgn);
  for (int c0=0;c0<numel;c0+=size[0]+1)
    sgn*=tmp.data[c0];
  return sgn;
}

template<class type>
tensor<type> tensor<type>::lu(tensor<int>& ind,type& sgn) {
  assert(ndims==2);
  assert(size[0]==size[1]);
  int i,imax,j,k;
  int n=size[0];
  type big,dum,sum,temp;
  tensor<type> ret(*this);
  tensor<type> vv(1,n);
  sgn=1;
  ind=tensor<int>(1,n);
  for (i=0;i<n;++i) {
    big=0;
    for (j=0;j<n;++j)
      if (big<(temp=abs(ret(i,j))))
        big=temp;
    if (abs(big)<TENSOR_EPS) { // THIS IS A HACK TO MAKE DET WORK
      tensor<type>::valid=0;
      return tensor<type>(ndims,size);
    }
    vv(i)=1/big;
  }
  for (j=0;j<n;++j) {
    for (i=0;i<j;++i) {
      sum=ret(i,j);
      for (k=0;k<i;++k)
        sum-=ret(i,k)*ret(k,j);
      ret(i,j)=sum;
    }
    big=0;
    for (i=j;i<n;++i) {
      sum=ret(i,j);
      for (k=0;k<j;++k)
        sum-=ret(i,k)*ret(k,j);
      ret(i,j)=sum;
      if (big<=(dum=vv(i)*abs(sum))) {
        big=dum;
        imax=i;
      }
    }
    if (j!=imax) {
      for (k=0;k<n;++k) {
        dum=ret(imax,k);
        ret(imax,k)=ret(j,k);
        ret(j,k)=dum;
      }
      sgn=-sgn;
      vv(imax)=vv(j);
    }
    ind(j)=imax;
    //if (abs(ret(j,j))<TENSOR_EPS) ret(j,j)=tiny
    if (j!=n-1) {
      dum=1/ret(j,j);
      for (i=j+1;i<n;++i)
        ret(i,j)*=dum;
    }
  }
  return ret;
}

template<class type>
tensor<type> tensor<type>::lu(tensor<int>& ind,tensor<type> rhs) {
  assert(rhs.ndims==1);
  assert(rhs.numel==size[0]);
  int i,ii=0,ip,j;
  int n=size[0];
  type sum;
  tensor<type> ret(rhs);
  for (i=0;i<n;++i) {
    ip=ind(i);
    sum=ret(ip);
    ret(ip)=ret(i);
    if (ii)
      for (j=ii-1;j<i;++j)
        sum-=data[i+j*n]*ret(j);
    else
      if (TENSOR_EPS<abs(sum))
        ii=i+1;
    ret(i)=sum;
  }
  for (i=n-1;i>=0;i--) {
    sum=ret(i);
    for (j=i+1;j<n;++j)
      sum-=data[i+j*n]*ret(j);
    ret(i)=sum/data[i+i*n];
  }
  return ret;
}

template<class type>
tensor<type> tensor<type>::qr(tensor<type>& ret) {
  ret=*this;
  tensor<type> qor=eye(size[0]);
  tensor<type> col(2,size[0],1);
  tensor<type> row(2,1,size[0]);
  tensor<type> tmp;
  tensor<type> prj(1,size[0]);
  for (int c0=0;c0<min(size[0]-1,size[1]);++c0) {
    prj.in(c0,&ret.data[c0+c0*size[0]],size[0]-c0);
    prj.data[c0]-=prj.norm();
    prj=prj.normal(); // TODO, normal() decides about the numerics in qr !?!?
    col.in(prj.data);
    col*=2;
    tmp=eye(size[0])-col*row.in(prj.data);
    ret=tmp*ret;
    qor=tmp*qor;
    prj=0;
  }
  return qor;
}

template<class type>
tensor<type> tensor<type>::schur(tensor<type>& ret) { // sor*ret*sor^t=*this (confirmed)
  assert(ndims==2);
  assert(size[0]==size[1]);
  tensor<double> sor=eye(size[0]);
  tensor<double> qor;
  tensor<double> tmp;
  tensor<double> cpy;
  ret=*this;
  for (int c0=0;c0<100;++c0) {
    cpy=ret.diag();
    qor=ret.qr(tmp);
    ret=qor*ret*qor.transpose();
    sor=qor*sor;
    valid=1;
    for (int c1=0;c1<size[1]&&valid;++c1)
      for (int c2=c1+1;c2<size[0]&&valid;++c2)
        valid&=abs(ret(c2,c1))<TENSOR_SCHUR; // values below diagonal are null
    if (valid)
      break;
  }
  assert(sor.transpose()*ret*sor==*this); // ok, because transpose is returned
  return sor.transpose();
}

#define  TENSOR_EIG_ROTATE(A,i,j,k,l)     {                                  \
                                            tg=A.data[i+j*n];                \
                                            th=A.data[k+l*n];                \
                                            A.data[i+j*n]=tg-s*(th+tg*tau);  \
                                            A.data[k+l*n]=th+s*(tg-th*tau);  \
                                          }

template<class type>
tensor<type> tensor<type>::eig(tensor<type>& v) { // v*diag(d)*v.transpose() == *this
  assert(ndims==2);
  assert(size[0]==size[1]);
  assert(*this==transpose());
  int i,j,ip,iq;
  type tresh,theta,tau,t,sm,s,h,g,c,tg,th;
  int n=size[0];
  v=eye(n);
  tensor<type> a=*this;
  tensor<type> d=diag();
  tensor<type> b=diag();
  tensor<type> z(1,size[0]);
  valid=1;
  for (i=0;i<50;++i) {
    sm=0;
    for (ip=0;ip<n-1;++ip)
      for (iq=ip+1;iq<n;++iq)
        sm+=abs(a(ip,iq));
    if (sm<TENSOR_EIG) { // sm == 0.0
      tensor<int> ind;
      d=d.sort(TENSOR_SORT_CBA,ind);
      v=v(tensor<int>::index(1,n),ind);
      assert( v.transpose()*v == tensor<double>::eye(n) );
      assert( v*v.transpose() == tensor<double>::eye(n) );
      assert( v*tensor<double>::diag(d)*v.transpose() == *this );
      return d;
    }
    tresh=(i<3)*0.2*sm/numel;
    for (ip=0;ip<n-1;++ip) {
      for (iq=ip+1;iq<n;++iq) {
        g=100*abs(a(ip,iq));
        if (i>3&&g<=TENSOR_NO_DIVISION*abs(d(ip))
               &&g<=TENSOR_NO_DIVISION*abs(d(iq)))
          a(ip,iq)=0;
        else
          if (tresh<abs(a(ip,iq))) {
            h=d(iq)-d(ip);
            if (g<=TENSOR_NO_DIVISION*abs(h))
              t=a(ip,iq)/h;
            else {
              theta=0.5*h/a(ip,iq);
              t=1/(abs(theta)+sqrt(1+theta*theta));
              if (theta<0)
                t=-t;
            }
            c=1/sqrt(1+t*t);
            s=t*c;
            tau=s/(1+c);
            h=t*a(ip,iq);
            z(ip)-=h;
            z(iq)+=h;
            d(ip)-=h;
            d(iq)+=h;
            a(ip,iq)=0;
            for (j=0;j<ip;++j)
              TENSOR_EIG_ROTATE(a,j,ip,j,iq)
            for (j=ip+1;j<iq;++j)
              TENSOR_EIG_ROTATE(a,ip,j,j,iq)
            for (j=iq+1;j<n;++j)
              TENSOR_EIG_ROTATE(a,ip,j,iq,j)
            for (j=0;j<n;++j)
              TENSOR_EIG_ROTATE(v,j,ip,j,iq)
            ++valid;
          }
      }
    }
    for (ip=0;ip<n;++ip) {
      b(ip)+=z(ip);
      d(ip)=b(ip);
      z(ip)=0;
    }
  }
  valid=0;
  assert(0);
  return d;
}

// right now:   w.transpose()*diag(dia)*v.transpose()==*this  (see assert below)
template<class type>
tensor<type> tensor<type>::svd(tensor<type>& w,tensor<type>& v) {
  assert(ndims==2);
  assert(size[1]<=size[0]);
  int flag;
  int i,its,j,jj,k,l,nm;
  type anorm,c,f,g,h,s,scale,x,y,z;
  int m=size[0];
  int n=size[1];
  w=*this;
  tensor<type> rv1(1,n);
  tensor<type> dia(1,n);     // w is output vector as diagonal of W
  v=tensor<type>(2,n,n);   // v is output matrix as V
  valid=1;
  // householder reduction to bidiagonal form
  g=scale=anorm=0;
  for (i=0;i<n;i++) {
    l=i+2;
    rv1(i)=scale*g;
    g=s=scale=0;
    if (i<m) {
      for (k=i;k<m;k++)
        scale+=abs(w(k,i));
      if (TENSOR_NO_DIVISION<abs(scale)) {
        for (k=i;k<m;k++) {
          w(k,i)/=scale;
          s+=w(k,i)*w(k,i);
        }
        f=w(i,i);
        g=-math::Sign(sqrt(s),f);
        h=f*g-s;
        w(i,i)=f-g;
        for (j=l-1;j<n;j++) {
          for (s=0,k=i;k<m;k++)
            s+=w(k,i)*w(k,j);
          f=s/h;
          for (k=i;k<m;k++)
            w(k,j)+=f*w(k,i);
        }
        for (k=i;k<m;k++)
          w(k,i)*=scale;
      }
    }
    dia(i)=scale*g;    
    g=s=scale=0;    
    if (i+1<=m&&i+1!=n) { // 3rd edition
      for (k=l-1;k<n;k++)
        scale+=abs(w(i,k));
      if (TENSOR_NO_DIVISION<abs(scale)) {
        for (k=l-1;k<n;k++) {
          w(i,k)/=scale;
          s+=w(i,k)*w(i,k);          
        }
        f=w(i,l-1);
        g=-math::Sign(sqrt(s),f);
        h=f*g-s;
        w(i,l-1)=f-g;        
        for (k=l-1;k<n;k++)
          rv1(k)=w(i,k)/h;        
        for (j=l-1;j<m;j++) {
          for (s=0,k=l-1;k<n;k++)
            s+=w(j,k)*w(i,k);          
          for (k=l-1;k<n;k++)
            w(j,k)+=s*rv1(k);
        }
        for (k=l-1;k<n;k++)
          w(i,k)*=scale;
      }
    }
    anorm=max(anorm,(abs(dia(i))+abs(rv1(i))));
  }
  // accumulation of right hand transformations
  for (i=n-1;i>=0;i--) {
    if (i<n-1) {
      if (TENSOR_NO_DIVISION<abs(g)) {
        for (j=l;j<n;j++)
          v(j,i)=w(i,j)/w(i,l)/g;
        for (j=l;j<n;j++) {
          for (s=0,k=l;k<n;k++)
            s+=w(i,k)*v(k,j);
          for (k=l;k<n;k++)
            v(k,j)+=s*v(k,i);
        }
      }
      for (j=l;j<n;j++)
        v(i,j)=v(j,i)=0;
    }
    v(i,i)=1;
    g=rv1(i);
    l=i;
  }
  for (i=min(m,n)-1;i>=0;i--) {
    l=i+1;
    g=dia(i);
    for (j=l;j<n;j++)
      w(i,j)=0;
    if (TENSOR_NO_DIVISION<abs(g)) {
      g=1/g;
      for (j=l;j<n;j++) {
        for (s=0,k=l;k<m;k++)
          s+=w(k,i)*w(k,j);
        f=s/w(i,i)*g;
        for (k=i;k<m;k++)
          w(k,j)+=f*w(k,i);
      }
      for (j=i;j<m;j++)
        w(j,i)*=g;
    } else
      for (j=i;j<m;j++)
        w(j,i)=0;
    w(i,i)+=1;
  }
  // diagonalization of the bidiagonal form
  for (k=n-1;k>=0;k--) {
    for (its=0;its<=TENSOR_SVD;its++) {
      flag=1;
      for (l=k;l>=0;l--) {
        nm=l-1;
        if (l==0||abs(rv1(l))<TENSOR_NO_DIVISION) { // 3rd edition
          flag=0;
          break;
        }
        if (abs(dia(nm))<TENSOR_NO_DIVISION)
          break;
      }
      if (flag) {
        c=0;
        s=1;
        for (i=l;i<k+1;i++) { // 3rd edition
          f=s*rv1(i);
          rv1(i)=c*rv1(i);
          if (abs(f)<TENSOR_NO_DIVISION)
            break;
          g=dia(i);
          h=sqrt(f*f+g*g);
          dia(i)=h;
          h=1/h;
          c=g*h;
          s=-f*h;
          for (j=0;j<m;j++) {
            y=w(j,nm);
            z=w(j,i);
            w(j,nm)=y*c+z*s;
            w(j,i)=z*c-y*s;
          }
        }
      }
      z=dia(k);
      if (l==k) {
        if (z<0) {
          dia(k)=-z;
          for (j=0;j<n;j++)
            v(j,k)=-v(j,k);
        }
        break;
      }
      valid&=its<TENSOR_SVD;
      x=dia(l);
      nm=k-1;
      y=dia(nm);
      g=rv1(nm);
      h=rv1(k);
      f=((y-z)*(y+z)+(g-h)*(g+h))/(2*h*y);
      g=sqrt(f*f+1);
      f=((x-z)*(x+z)+h*((y/(f+math::Sign(g,f)))-h))/x;
      c=s=1;
      for (j=l;j<=nm;j++) {
        i=j+1;
        g=rv1(i);
        y=dia(i);
        h=s*g;
        g=c*g;
        z=sqrt(f*f+h*h);
        rv1(j)=z;
        c=f/z;
        s=h/z;
        f=x*c+g*s;
        g=g*c-x*s;
        h=y*s;
        y*=c;
        for (jj=0;jj<n;jj++) {
          x=v(jj,j);
          z=v(jj,i);
          v(jj,j)=x*c+z*s;
          v(jj,i)=z*c-x*s;
        }
        z=sqrt(f*f+h*h);
        dia(j)=z;
        if (TENSOR_NO_DIVISION<abs(z)) {
          z=1/z;
          c=f*z;
          s=h*z;
        }
        f=c*g+s*y;
        x=c*y-s*g;
        for (jj=0;jj<m;jj++) {
          y=w(jj,j);
          z=w(jj,i);
          w(jj,j)=y*c+z*s;
          w(jj,i)=z*c-y*s;
        }
      }
      rv1(l)=0;
      rv1(k)=f;
      dia(k)=x;
    }
  }
  tensor<int> ind;
  dia=dia.sort(TENSOR_SORT_CBA,ind);
  v=v(tensor<int>::index(1,n),ind); // apply sorting
  w=w(tensor<int>::index(1,m),ind).transpose();
  if (!(w.transpose()*diag(dia)*v.transpose()==*this)) {
    tensor<type> err=w.transpose()*diag(dia)*v.transpose()-*this;    
    printf("svd delta %10.2e\n",err.norm());
    //printf("%s\n\n ",toString());
    //printf("%s\n\n ",(w.transpose()*diag(dia)*v.transpose()).toString());
    //exit(1);
  }
#ifndef TENSOR_SVD_NDEBUG
  assert(w.transpose()*diag(dia)*v.transpose()==*this);
#endif
  return dia;
}

template<class type>
tensor<type> tensor<type>::polar(tensor<type>& pol) {
  tensor<type> w,v;
  tensor<type> dia=svd(w,v);
  tensor<type> ret=v*w;
  pol=v*diag(dia)*v.transpose();
  assert(ret.transpose()*pol==*this);
  return ret;
}

// if *this=W'*D*V'  then  this->inverse() = V*D^(-1)*W
template<class type>
tensor<type> tensor<type>::inverse(void) {
  return inverse(size[1]);
}

template<class type>
tensor<type> tensor<type>::inverse(int num) {
  int c0,c1,cnt=0;
  type fac;
  tensor<type> w,v;
  tensor<type> val=svd(w,v);
  valid=0;
  for (c0=0;c0<num;++c0)
    if (precision<val.data[c0]) {
      ++valid;
      fac=1/val.data[c0];
      for (c1=0;c1<v.size[0];c1++)
        v.data[cnt++]*=fac;
    } else {
      printf("warning: rank %i smaller than expected %i.\n",c0,num);
      break;
    }
  for (c0=cnt;c0<v.numel;++c0)
    v.data[c0]=0;
  return v*w;
}

// if this=U*D*V'  then  pinv(this)=V*D^(-1)*U'
//template<class type>
//tensor<type> tensor<type>::null(void) {
//  //tensor<type> dia,ort;
//  //tensor<type> urt=svd(dia,ort);
//  //for (int c1=0;c1<dia.numel;++c1)
//  //  if (TENSOR_NO_DIVISION<dia.data[c1])
//  //    dia.data[c1]=1/dia.data[c1];
//  return *this;
//}

template<class type>
int tensor<type>::rank(void) {
  int c0;
  tensor<type> dia,w,v;
  if (size[0]<size[1])
    dia=transpose().svd(w,v);
  else
    dia=svd(w,v);
  for (c0=0;c0<dia.numel;c0++)
    if (dia.data[c0]<=precision)
      break;
  return c0;
}

template<class type>
int tensor<type>::newton(type (*func)(tensor<type>),tensor<type> g,tensor<type> p,type stpmax) {
  // F(x)=0  ... func= F.F/2    g= d(func)    p= - dF^-1 . F
  type a,alam=1,alam2=0,alamin,b,disc,f,f2=0;
  type rhs1,rhs2,slope,sum,temp,test=0,tmplam;
  type fold=func(*this);
  sum=p.norm();
  if (sum>stpmax)
    p*=stpmax/sum;
  slope=g|p;
  if (0<=slope)
    return TENSOR_NEWTON_CHECK; // can be a zero like 1-cos(x)
  for (int c0=0;c0<numel;++c0) {
    temp=abs(p(c0))/max(abs(data[c0]),1);
    if (test<temp)
      test=temp;
  }
  alamin=TENSOR_EPS/test;
  tensor<type> x;
  while (1) {
    x=*this+p*alam;
    f=func(x);
    if (alam<alamin)
      return TENSOR_NEWTON_CHECK; // still need to check convergence !!!
    else
      if (f<=fold+1e-4*alam*slope) { // MAGIC CONST
        *this=x;
        return TENSOR_NEWTON_CONT; 
      } else {
        if (alam==1) // TODO bad style
          tmplam=-slope/(2*(f-fold-slope));
        else {
          rhs1=f-fold-alam*slope;
          rhs2=f2-fold-alam2*slope;
          a=(rhs1/(alam*alam)-rhs2/(alam2*alam2))/(alam-alam2);
          b=(-alam2*rhs1/(alam*alam)+alam*rhs2/(alam2*alam2))/(alam-alam2);
          if (a==0)
            tmplam=-slope/(2*b);
          else {
            disc=b*b-3*a*slope;
            if (disc<0)
              tmplam=0.5*alam;
            else
              if (b<=0)
                tmplam=(-b+sqrt(disc))/(3*a);
              else
                tmplam=-slope/(b+sqrt(disc));            
          }
          if (0.5*alam<tmplam)
            tmplam=0.5*alam;
        }
      }
    alam2=alam; // TODO check parentheses!
    f2=f;
    alam=max(tmplam,0.1*alam); 
  }
}

// * each dimension is parametrized as [0...1]
// * for vector(4), itp=2/3 => data[2]
// * also works for type==uchar
// TODO test with size[.]==1
template<class type>
type tensor<type>::interp(tensor<double> itp) {
  assert(itp.numel==ndims);
  int c0,id0,id1,adr;
  double fac;
  tensor<int> cum=cumsize();
  tensor<int> idx(2,2,ndims);
  tensor<double> rem(2,2,ndims);
  for (c0=0;c0<ndims;++c0) {
    itp(c0)=min(max(0,itp(c0)),1);
    fac=itp(c0)*(size[c0]-1);
    id0=fac;
    id1=id0+1;
    rem(0,c0)=id1-fac;
    rem(1,c0)=1-rem(0,c0);
    idx(0,c0)=min(max(0,id0),size[c0]-1);
    idx(1,c0)=min(max(0,id1),size[c0]-1);
  }
  uint msk,pwr=1<<ndims;
  double sum=0;
  for (uint c1=0;c1<pwr;++c1) {
    msk=c1;
    fac=1;
    adr=0;
    for (c0=0;c0<ndims;++c0) {
      fac*=rem(msk&1,c0);
      adr+=idx(msk&1,c0)*cum(c0);
      msk>>=1;
    }
    sum+=fac*(double)data[adr];
  }
  return sum;
}

//template<class type>
//type tensor<type>::interp(tensor<double> itp,tensor<int>& idx) {
//  assert(itp.numel==ndims);
//  int c0,id0,id1,adr;
//  double fac;
//  tensor<int> cum=cumsize();
//  idx=tensor<int>(2,ndims,2);
//  tensor<double> rem(2,ndims,2);
//  for (c0=0;c0<ndims;++c0) {
//    itp(c0)=min(max(0,itp(c0)),1);
//    fac=itp(c0)*(size[c0]-1);
//    id0=fac;
//    id1=id0+1;
//    rem(c0,0)=id1-fac;
//    rem(c0,1)=1-rem(c0,0);
//    idx(c0,0)=min(max(0,id0),size[c0]-1);
//    idx(c0,1)=min(max(0,id1),size[c0]-1);
//  }
//  uint msk,pwr=1<<ndims;
//  double sum=0;
//  for (uint c1=0;c1<pwr;++c1) {
//    msk=c1;
//    fac=1;
//    adr=0;
//    for (c0=0;c0<ndims;++c0) {
//      fac*=rem(c0,msk&1);
//      adr+=idx(c0,msk&1)*cum(c0);
//      msk>>=1;
//    }
//    sum+=fac*(double)data[adr];
//  }
//  return sum;
//}
